<?php


final class AmazonPay
{
    private static array $config;
    private static string $payload;
    private static string $sandbox;
    private static string $site;
    private static string $environment;
    private static AmazonPay $instance;
    private static $client;


    private function __construct(){}

    private function __clone() {}

    public static function getInstance(): self {
        if(!isset(self::$instance)){
            self::$instance = new self();
        }

        self::init();

        return self::$instance;
    }

    private static function init(): void
    {
        self::$site = explode('_', SITE_INSTANCE)[1];
        self::$environment = Config::getConfigParam('amazon_mode');
        self::$sandbox = self::$environment === 'sandbox' ? '_sandbox' : '';
        $private_key_path = str_replace('{site}', self::$site, Config::getConfigParam('amazon_private_key'.self::$sandbox));

        self::$config = array(
            'public_key_id' => Config::getConfigParam('amazon_public_key_id'.self::$sandbox),
            'private_key' => $private_key_path,
            'region' => 'US'
        );

        if(self::$environment === 'sandbox'){
            self::$config['sandbox'] = true;
        }

        self::$client = new \Amazon\Pay\API\Client(self::$config);

        self::$payload = '{"scopes": ["name", "email", "phoneNumber", "billingAddress"], "storeId":"' . Config::getConfigParam('amazon_store_id'.self::$sandbox) . '","webCheckoutDetails":{"checkoutReviewReturnUrl":"' . str_replace('{site}', self::$site, Config::getConfigParam('amazon_return_url'.self::$sandbox)) . '"}}';
    }

    protected function chargeAmount(): float
    {
        $orders = Transaction::retrieve()->getOrderProperties();
        $total = 0.00;
        foreach ($orders as $order) {
            $total += $order['total'];
        }
        return round($total, 2);
    }

    public function createCheckoutButton(): ?string
    {
        try {
            $signature = self::$client->generateButtonSignature(self::$payload);
            if ($signature) {
                return json_encode([
                    'merchantId' => Config::getConfigParam('amazon_merchant_id'.self::$sandbox),
                    'publicKeyId' => Config::getConfigParam('amazon_public_key_id'.self::$sandbox),
                    'amount' => $this->chargeAmount(),
                    'payload' => self::$payload,
                    'signature' => $signature
                ]);
            }
        } catch (\Exception $e) {
            error_log($e);
        }

        return json_encode(['error' => '']);
    }

    protected function getCheckoutSessionId(): string
    {
        return $_SESSION['AMAZON_CHECKOUT_SESSION_ID'];
    }

    protected function setCheckoutSessionId(string $checkoutSessionId): void
    {
        $_SESSION['AMAZON_CHECKOUT_SESSION_ID'] = $checkoutSessionId;
    }

    private function getMerchantReferenceId(): int
    {
        $merchantReferenceId = -1;
        $options = ['headers' => ['Host' => 'bloomstoday.net'], 'verify' => false];
        try {
            $client = new Client();
            $request = $client->request('GET', "https://184.106.216.224/submit_order.php", $options);
            $response = $request->getBody()->getContents();
            if ($response) {
                $response = unserialize($response);
                $merchantReferenceId = $response['order_number'];
            }
        } catch (GuzzleException $e) {
            error_log($e);
        }
        return $merchantReferenceId;
    }

    public function updateCheckoutSession(): ?string
    {
        $payload = array(
            "webCheckoutDetails" => [
                "checkoutResultReturnUrl" => str_replace('{site}', self::$site, Config::getConfigParam('amazon_return_url'.self::$sandbox)).'?complete=true'
            ],
            'paymentDetails' => array(
                'paymentIntent' => 'Authorize',
                'canHandlePendingAuthorization' => false,
                'chargeAmount' => array(
                    'amount' => $this->chargeAmount(),
                    'currencyCode' => 'USD'
                ),
            ),
            'merchantMetadata' => array(
                'merchantReferenceId' => $this->getMerchantReferenceId(),
                'merchantStoreName' => Config::getConfigParam('site_name'),
                'noteToBuyer' => 'Thank you for your order!'
            )
        );

        try {
            $result = self::$client->updateCheckoutSession($this->getCheckoutSessionId(), $payload);
            error_log(json_encode($result));
            if ($result['status'] === 200) {
                $response = json_decode($result['response'], true);
                $amazonPayRedirectUrl = $response['webCheckoutDetails']['amazonPayRedirectUrl'];
                return json_encode(['processing_url' => $amazonPayRedirectUrl]);
            }
        } catch (\Exception $e) {
            error_log($e);
        }
        return null;
    }

    public function completeCheckoutSession(): ?string {
        $payload = array(
            'chargeAmount' => array(
                'amount' => $this->chargeAmount(),
                'currencyCode' => 'USD'
            )
        );

        try {
            $result = self::$client->completeCheckoutSession($this->getCheckoutSessionId(), $payload);
            error_log(json_encode($result));
            if ($result['status'] === 200) {
                $response = json_decode($result['response'], true);
                return $response['chargeId'];
            }
        } catch (\Exception $e) {
            error_log($e);
        }
        return null;
    }

    public function captureCharge(string $chargeId): bool {
        $success = false;
        $payload = array(
            'captureAmount' => array(
                'amount' => $this->chargeAmount(),
                'currencyCode' => 'USD'
            )
        );
        $headers = array('x-amz-pay-idempotency-key' => uniqid());
        try {
            $result = self::$client->captureCharge($chargeId, $payload, $headers);
            error_log(json_encode($result));
            $success = $result['status'] === 200;
        } catch (\Exception $e) {
            error_log($e);
        }

        return $success;
    }


    public function getCheckoutSession(string $checkoutSessionId): ?array
    {
        try {
            $this->setCheckoutSessionId($checkoutSessionId);
            $result = self::$client->getCheckoutSession($checkoutSessionId);
            error_log(json_encode($result));
            if ($result['status'] === 200) {
                $response = json_decode($result['response'], true);
                $buyerName = $response['buyer']['name'];
                $firstname = explode(' ', $buyerName)[0];
                $lastname = explode(' ', $buyerName)[1];
                $phone = $response['buyer']['phoneNumber'];
                $buyerEmail = $response['buyer']['email'];
                $addrLine1 = $response['billingAddress']['addressLine1'];
                $addrLine2 = $response['billingAddress']['addressLine2'];
                $zip = $response['billingAddress']['postalCode'];
                return [
                    'firstname' => $firstname,
                    'lastname' => $lastname,
                    'address1' => $addrLine1,
                    'address2' => $addrLine2,
                    'zip' => $zip,
                    'phone' => $phone,
                    'email' => $buyerEmail
                ];
            }
        } catch (\Exception $e) {
            // handle the exception
            error_log($e);
        }
        return null;
    }


}