<?php

namespace CuteHttp;

use \CuteLogger\Logger;

class Client
{
    private $handle;
    //private StreamInterface $stream;
    private ?string $base_uri = null;
    private ?string $verify = null;
    private ?string $body = null;
    private ?string $cookie = null;
    private string $method;
    private ?string $auth = null;
    private ?string $query = null;
    private ?string $url = null;
    private array $headers = [];
    private array $options = [];
    private string $errno;
    private string $error;


    public function __construct(array $options)
    {
        $this->handle = curl_init();
        $this->mergeOptions($options);
    }

    private function mergeOptions(array $options): void
    {
        $this->options = array_merge($this->options, $options);
    }

    public function request($method, $uri, array $options): bool|string|null
    {
        try {
            $this->method = strtolower($method);
            $this->mergeOptions($options);
            $this->setOptions($this->options);
            $this->build_uri($uri);

            if ($this->method === 'post' || $this->method === 'put' || $this->method === 'patch') {
                curl_setopt($this->handle, CURLOPT_POST, true);
                curl_setopt($this->handle, CURLOPT_POSTFIELDS, $this->body);
            }

            return $this->execute();

        } catch (\Exception $e) {
            Logger::Exception($e);
        }
        return false;
    }

    private function execute()
    {
        try {
            $response = curl_exec($this->handle);
            $this->errno = curl_errno($this->handle);
            $this->error = curl_error($this->handle);
            curl_close($this->handle);
            return str_contains($response, '&') ? $this->parseQueryString($response) : $response;
        } catch (\Exception $e) {
            Logger::Exception($e);
        }
    }

    public function getError(): string
    {
        return "$this->errno: $this->error";
    }

    public function addHeader($key, $value): void
    {
        $this->headers[] = "$key: $value";
    }

    private function setOptions($options): void
    {
        if (array_key_exists('base_uri', $options)) {
            $this->base_uri = $options['base_uri'];
        }

        if (array_key_exists('query', $options)) {
            $this->query = http_build_query($options['query']);
        }

        if (array_key_exists('headers', $options)) {
            foreach ($options['headers'] as $key => $val) {
                if (!is_array($val)) {
                    if (is_bool($val)) {
                        $this->addHeader($key, $val ? '1' : '0');
                    } else {
                        $this->addHeader($key, $val);
                    }
                }
            }
        }

        if (array_key_exists('body', $options)) {
            $this->body = $options['body'];
        }

        if (array_key_exists('form_params', $options)) {
            //$this->headers[]='Content-Type: application/x-www-form-urlencoded';
            $this->body = http_build_query($options['form_params']);
        }


        $this->headers[] = 'Cache-Control: no-cache';
        $this->headers[] = 'Language: en_US';

        if ($this->body) {
            $this->headers[] = 'Content-length: ' . strlen($this->body);
        }

        if (isset($options['headers'])) {
            if (array_key_exists('verify', $options['headers'])) {
                $this->verify = $options['headers']['verify'] ? 'true' : 'false';
            } else if (array_key_exists('verify', $options)) {
                $this->verify = $options['verify'] ? 'true' : 'false';
            }

            if ($this->verify) {
                curl_setopt($this->handle, CURLOPT_SSL_VERIFYPEER, $this->verify === 'true' ? 1 : 0);
                curl_setopt($this->handle, CURLOPT_SSL_VERIFYHOST, $this->verify === 'true' ? 2 : 0);
            }

            if (array_key_exists('cookie', $options['headers'])) {
                $this->cookie = 'Cookie: ' . $options['headers']['cookie'][0] . '=' . $options['headers']['cookie'][1];
            } else if (array_key_exists('cookie', $options)) {
                $this->cookie = 'Cookie: ' . $options['cookie'][0] . '=' . $options['cookie'][1];
            }

            if ($this->cookie) {
                $this->headers[] = $this->cookie;
            }

            if (array_key_exists('auth', $options['headers'])) {
                $user = $options['headers']['auth'][0];
                $pass = $options['headers']['auth'][1];
                $this->auth = "Authorization: Basic " . base64_encode("$user:$pass");
            } else if (array_key_exists('auth', $options)) {
                $user = $options['auth'][0];
                $pass = $options['auth'][1];
                $this->auth = "Authorization: Basic " . base64_encode("$user:$pass");
            }

            if ($this->auth) {
                $this->headers[] = $this->auth;
            }
        }

        if (isset($options['timeout'])) {
            curl_setopt($this->handle, CURLOPT_CONNECTTIMEOUT, $options['timeout']);
            curl_setopt($this->handle, CURLOPT_TIMEOUT, $options['timeout']);
        }

        curl_setopt($this->handle, CURLOPT_FOLLOWLOCATION, 1);
        curl_setopt($this->handle, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($this->handle, CURLOPT_HEADER, 0);
        curl_setopt($this->handle, CURLOPT_HTTPHEADER, $this->headers);
    }

    private function build_uri($uri): void
    {
        if ($this->base_uri) {
            if (str_ends_with($this->base_uri, '/')) {
                $this->url = substr($this->base_uri, 0, -1);
            } else {
                $this->url = $this->base_uri;
            }
        }

        if ($uri) {
            if (str_starts_with($uri, '/')) {
                $this->url .= '/' . substr($uri, 1);
            }
        }

        if ($this->query) {
            if (str_starts_with($uri, '?')) {
                $this->query = substr($this->query, 1);
            }
            if ($this->method === 'get') {
                $this->url .= '?' . $this->query;
            }
        }

        curl_setopt($this->handle, CURLOPT_URL, $this->url);
    }


    private function parseQueryString(string $string): string {
        $result = "";
        $array = [];

        if($params = explode('&', $string)){
            foreach($params as $param){
                $inner = explode('=', $param);
                $array[$inner[0]] = urldecode($inner[1]);
            }
            $result = isset($array) ? json_encode($array) : "";
        }

        return $result;
    }


}