<?php

use \CuteLogger\Logger;

class Database
{
    private PDO $pdo;
    private string $sql;
    private array $params;
    private static string $host;
    private static string $user;
    private static string $pass;
    private static string $dbname;
    private static PDOStatement $statement;
    private static ?PDO $instance = null;
    private array $_fields = array();
    private array $_values = array();


    /**
     * @param $host
     * @param $user
     * @param $pass
     * @param $dbname
     */
    public function __construct($host, $user, $pass, $dbname)
    {
        try {
            self::$host = $host;
            self::$user = $user;
            self::$pass = $pass;
            self::$dbname = $dbname;
            $this->pdo = self::getInstance();
        } catch (PDOException $e) {
            Logger::Exception($e);
        }
    }

    /**
     * @return PDO|null
     */
    public static function getInstance(): ?PDO
    {
        if (empty(self::$instance)) {
            try {
                $dsn = "mysql:host=" . self::$host . ";dbname=" . self::$dbname;
                $options = array(PDO::ATTR_ERRMODE => PDO::ERRMODE_EXCEPTION);
                self::$instance = new PDO($dsn, self::$user, self::$pass, $options);
                self::$instance->setAttribute(PDO::MYSQL_ATTR_USE_BUFFERED_QUERY, false);
            } catch (PDOException $e) {
                Logger::Exception($e);
            }
        }
        return self::$instance;
    }

    /**
     * @return null
     */
    public static function close(): null
    {
        return self::$instance = null;
    }

    /**
     * @param $query
     * @return $this
     */
    public function setQuery($query): self
    {
        $this->sql = $query;
        return $this;
    }


    /**
     * @return string
     */
    public function getQuery(): string
    {
        return $this->sql;
    }

    /**
     * @return string
     */
    public function getError(): string
    {
        return json_encode($this->pdo->errorInfo());
    }

    /**
     * @return string
     */
    public function getErrorCode(): string
    {
        return $this->pdo->errorCode();
    }

    /**
     * @return $this
     */
    public function params(): self
    {
        $args = func_get_args();
        foreach ($args as $arg) {
            $this->params[] = $this->escape($arg);
        }
        return $this;
    }

    public function query(): self
    {
        if ($this->sql) {
            try {
                self::$statement = $this->pdo->query($this->sql);
            } catch (PDOException $e) {
                Logger::Exception($e);
            }
        }
        return $this;
    }

    public function rowCount(): int
    {
        return self::$statement->rowCount();
    }

    public function execute(): self
    {
        if ($this->sql) {
            try {
                self::$statement = $this->pdo->prepare($this->sql);
                if (isset($this->params)) {
                    self::$statement->execute($this->params);
                } else {
                    self::$statement->execute();
                }
            } catch (PDOException $e) {
                Logger::Exception($e);
            }
        }

        return $this;
    }

    public function loadResult(): mixed
    {
        $result = null;
        try {
            self::$statement = $this->pdo->query($this->sql);
            $result = self::$statement->fetch(PDO::FETCH_OBJ);
            self::$statement->closeCursor();
        } catch (PDOException $e) {
            Logger::Exception($e);
        }
        return $result;
    }

    public
    function loadObject(): bool|array|null
    {
        $result = null;

        try {
            self::$statement = $this->pdo->query($this->sql);
            $result = self::$statement->fetchAll(PDO::FETCH_OBJ);
            self::$statement->closeCursor();
        } catch (PDOException $e) {
            Logger::Exception($e);
        }
        return $result;
    }

    /**
     * @return false|string
     */
    public
    function lastInsertId(): bool|string
    {
        try {
            return $this->pdo->lastInsertId();
        } catch (PDOException $e) {
            Logger::Exception($e);
        }
        return false;
    }

    public function quoteName(string $field): string
    {
        $string = str_replace(['`', '\"', '\''], '', $field);

        return "`$string`";
    }

    /**
     * @param string $text
     * @param bool $escape
     * @return string
     */
    public
    function quote(string $text, bool $escape = true): string
    {
        if (strlen($text) <= 0) return "''";
        $text = stripslashes($text);
        if($this->isArray($text)){
            return "'$text'";
        } else {
            return $escape ? $this->pdo->quote($this->escape($text)) : $this->pdo->quote($text);
        }
    }

    private function isArray(string $text): bool {
        return str_starts_with($text, '[') && str_ends_with($text, ']');
    }

    private function stripquotes(string $text): string{
        return str_replace(['"',"'"], "", $text);
    }


    /**
     * @param string $text
     * @return string
     */
    public
    function escape(string $text): string
    {
        $search = array("\\", "\x00", "\n", "\r", "'", '"', "\x1a"/*, ":", "?"*/);
        $replace = array("\\\\", "\\0", "\\n", "\\r", "\'", '\"', "\\Z"/*, "&#58;", "&#63;"*/);
        return str_replace($search, $replace, $text);
    }

    private function getTableColumns(string $table): array
    {
        $result = [];

        if ($fields = $this->setQuery("SHOW FULL COLUMNS FROM " . $this->quoteName($table))->loadObject()) {

            foreach ($fields as $field) {
                $result[$field->Field] = preg_replace('/[(0-9)]/', '', $field->Type);
            }
        }

        return $result;
    }

    private function setTableFieldsAndValues(string $table, &$object, bool $quoteValues = false): void{

        $tableColumns = $this->getTableColumns($table);

        unset($this->_fields, $this->_values);

        foreach (get_object_vars($object) as $k => $v) {

            // Skip columns that don't exist in the table.
            if (!array_key_exists($k, $tableColumns)) {
                continue;
            }

            // Only process non-null scalars.
            if (\is_array($v) || \is_object($v) || $v === null) {
                continue;
            }

            // Ignore null datetime fields.
            if ($tableColumns[$k] === 'datetime' && empty($v)) {
                continue;
            }

            // Ignore null integer fields.
            if (stristr($tableColumns[$k], 'int') !== false && $v === '') {
                continue;
            }

            // Prepare and sanitize the fields and values for the database query.
            $this->_fields[] = $this->quoteName($k);
            $this->_values[] = $quoteValues ? $this->quote($v) : $v;
        }
    }

    private function toValuePair(array $keys, array $values): array {
        $result = [];
        foreach(array_combine($keys, $values) as $key => $val) {
            $result [$key] = $val;
        }
        return $result;
    }

    public function updateObject(string $table, &$object, string $where): bool{
        $this->setTableFieldsAndValues($table, $object, false);
        $values = $this->toValuePair($this->_fields, $this->_values);
        $this->update($table)
            ->set($values)
            ->where($where)
            ->execute();

        return $this->rowCount() > 0;
    }

    public function insertObject(string $table, &$object): int
    {
        $this->setTableFieldsAndValues($table, $object, false);
        $this->insert($this->quoteName($table))
            ->columns($this->_fields)
            ->values($this->_values)
            ->execute();

        return $this->lastInsertId();
    }

    public function insert(string $table): self
    {
        $this->sql = "INSERT INTO " . $this->quoteName($table) . " ";

        return $this;
    }

    public function columns(array $columns): self
    {
        $values = [];

        foreach ($columns as $column) {
            $values [] = $this->quoteName($column);
        }

        $this->sql .= "(" . join(',', $values) . ") ";

        return $this;
    }

    public function values(array $values): self
    {
        $results = [];

        foreach ($values as $value) {
            $results [] = $this->quote($value); //$this->escape($value);
        }

        $this->sql .= "VALUES (" . join(',', $results) . ") ";

        return $this;
    }

    public
    function set(array $columns): self
    {
        $values = [];

        foreach ($columns as $key => $value) {
            $values[] = $this->quoteName($key) . '=' . $this->quote($value);
        }

        $this->sql .= "SET " . join(',', $values) . " ";

        return $this;
    }

    public
    function update(string $table): self
    {
        $this->sql = "UPDATE " . $this->quoteName($table) . " ";

        return $this;
    }

    public
    function delete(string $table): self
    {
        $this->sql = "DELETE FROM $table ";

        return $this;
    }

    public
    function select(string $column): self
    {
        $this->sql = "SELECT " . $column . " ";

        return $this;
    }

    public
    function from(string $table): self
    {
        $this->sql .= "FROM " . $table . " ";

        return $this;
    }

    public
    function where(string $where): self
    {
        $this->sql .= "WHERE $where ";

        return $this;
    }

    public function whereIn(string $column, array $values): self{

        $list = array();

        foreach ($values as $value){
            $list [] = $this->quote($value);
        }

        $this->sql .= "WHERE ". $this->quoteName($column). " IN (" . join(',', $list). ") ";

        return $this;
    }

    public
    function order(string $order, string $direction = "DESC"): self
    {
        $this->sql .= "ORDER BY $order $direction ";

        return $this;
    }

    public function join(string $type, string $table): self
    {
        $this->sql .= "$type JOIN $table ";
        
        return $this;
    }

    public function on(string $statement): self
    {
        $this->sql .= "ON ($statement) ";

        return $this;
    }

    public function group(string $statement): self
    {
        $this->sql .= "GROUP BY $statement ";

        return $this;
    }

    public function having(string $statement): self {
        $this->sql .= "HAVING $statement ";

        return $this;
    }

    public function limit(int $row_count, int $offset = 0): self
    {
        $this->sql .= "LIMIT $offset, $row_count";

        return $this;
    }
}